@@ -1912,11 +1912,26 @@ void GnuHashSection<E>::copy_buf(Context<E> &ctx) {
 
 template <typename E>
 std::string_view
-get_merged_output_name(Context<E> &ctx, std::string_view name, u64 flags) {
+get_merged_output_name(Context<E> &ctx, std::string_view name, u64 flags,
+                       i64 entsize, i64 addralign) {
   if (ctx.arg.relocatable && !ctx.arg.relocatable_merge_sections)
     return name;
   if (ctx.arg.unique && ctx.arg.unique->match(name))
     return name;
+
+  // GCC seems to create sections named ".rodata.strN.<mangled-symbol-name>.M".
+  // We want to eliminate the symbol name part from the section name.
+  if ((flags & SHF_STRINGS) && name.starts_with(".rodata.")) {
+    if (entsize == 1 && addralign == 1)
+      return ".rodata.str1.1";
+    if (entsize == 2 && addralign == 2)
+      return ".rodata.str2.2";
+    if (entsize == 4 && addralign == 4)
+      return ".rodata.str4.4";
+    return save_string(ctx,".rodata.str"s + std::to_string(entsize) + "." +
+                       std::to_string(addralign));
+  }
+
   return name;
 }
 
@@ -1932,8 +1947,9 @@ MergedSection<E>::MergedSection(std::string_view name, i64 flags, i64 type,
 template <typename E>
 MergedSection<E> *
 MergedSection<E>::get_instance(Context<E> &ctx, std::string_view name,
-                               i64 type, i64 flags, i64 entsize) {
-  name = get_merged_output_name(ctx, name, flags);
+                               i64 type, i64 flags,
+                               i64 entsize, i64 addralign) {
+  name = get_merged_output_name(ctx, name, flags, entsize, addralign);
   flags = flags & ~(u64)SHF_GROUP & ~(u64)SHF_COMPRESSED;
 
   auto find = [&]() -> MergedSection * {
@@ -2534,8 +2550,12 @@ void VerdefSection<E>::construct(Context<E> &ctx) {
   for (std::string_view verstr : ctx.arg.version_definitions)
     write(verstr, idx++, 0);
 
-  for (Symbol<E> *sym : std::span<Symbol<E> *>(ctx.dynsym->symbols).subspan(1))
-    ctx.versym->contents[sym->get_dynsym_idx(ctx)] = sym->ver_idx;
+  for (Symbol<E> *sym : std::span<Symbol<E> *>(ctx.dynsym->symbols).subspan(1)) {
+    i64 ver = sym->ver_idx;
+    if (ver == VER_NDX_UNSPECIFIED)
+      ver = VER_NDX_GLOBAL;
+    ctx.versym->contents[sym->get_dynsym_idx(ctx)] = ver;
+  }
 }
 
 template <typename E>
